''' Phylogenetic tree scoring models; intended to be coupled with the use of pytbeaglehon (BEAGLE) high-performance library. '''

# Date:   Jun 9 2014
# Author: Alex Safatli
# E-mail: safatli@cs.dal.ca

# High-Performance Numeric Computation Library (pandas)

import pandas as P

# pytbeaglehon coupling

pytbeaglehonEnabled = True
try:
    import pytbeaglehon.disc_state_cont_time_model as BEAGLEModels
    from pytbeaglehon.disc_char_type import DNAType, AAType
    from pytbeaglehon.tests.util import TreeForTesting as BEAGLETree
except: pytbeaglehonEnabled = False

# Exception Handling

class PhyloModelError(Exception):
    def __init__(self,v): self.value = v
    def __str__(self): return repr(self.value)

# Model a biological sequence as a discrete state model.

class DiscreteStateModel(object):
    
    ''' Initialize a discrete state model 
    for phylogenetic data. State frequencies and character
    time are determined from the given alignment object. '''
    
    def __init__(self,alignment):
        
        self.alignment    = alignment
        self.charType     = None
        self.stateList    = None
        self.seqMatrix    = None
        self.rawFreqMap   = None
        self.rawFreqs     = None
        self.stateFreqs   = None
        self._totalFreq   = 0.0
        if pytbeaglehonEnabled:
            self._determineCharType()
            self._transformAlignmentToMatrix()
            self._determineStateFreqs()
        
    # Internals
    
    def _determineCharType(self):
        
        ''' PRIVATE: Determine the character type intrinsic to the
        biological sequence alignment given to the instance. '''
        
        a = self.alignment
        if (a.getDim() == 4): self.charType = DNAType()
        elif (a.getDim() >= 20 and a.getDim() <= 21):
            self.charType = AAType()
        else: raise PhyloModelError(
            'Unrecognizable character type in alignment.')
    
    def _transformAlignmentToMatrix(self):
        
        ''' PRIVATE: Transform instance alignment to a matrix of 
        states as integers. '''
        
        lioflists = []
        strList   = map(str.upper,self.alignment.toStrList())
        translate = self.charType.to_indices
        for l in strList: lioflists.append(translate(l))
        self.stateList = lioflists
        self.seqMatrix = P.DataFrame(lioflists)
    
    def _determineStateFreqs(self):
        
        ''' PRIVATE: Compute the state frequencies for all states if
        character type is DNA. Otherwise, leave undefined. '''
    
        if self.charType.get_num_states() != 4: return # DNAType?
        self.rawFreqMap = {}
        
        # Get all frequencies from the sequence matrix dataframe.
        freqS = self.seqMatrix.stack().value_counts()
        
        # Get all present states that are found in the dataframe.
        present = freqS.keys()
        
        # Get all possible states.
        states = self.charType.get_states()
        
        # Enumerate over all possible states.
        for state in states:
            i = self.charType.to_indices([state])[0]
            if i in present:
                self.rawFreqMap[state] = freqS.get(i)
            else: self.rawFreqMap[state] = 0
        self.rawFreqs = self.rawFreqMap.items() # Flatten.
        
        # Calculate actual state frequencies.
        total = float(sum([i[1] for i in self.rawFreqs]))
        self._totalFreq = total
        self.stateFreqs = map(lambda x:x[1]/total,self.rawFreqs)
    
    # Accessors    
    
    def getAlignment(self): return self.alignment
    def getAlignmentAsStateList(self): return self.stateList
    def getSequenceMatrix(self): return self.seqMatrix
    def getCharType(self): return self.charType
    def getStateFreqs(self): return self.stateFreqs
    def getRawStateFreqs(self): return self.rawFreqs
    def getRawStateFreqsAsList(self): return self.getRawStateFreqs()
    def getRawStateFreqsAsDict(self): return self.rawFreqMap
    
    def getFrequencyOfState(self,i):
        if self.rawFreqMap:
            if i in self.rawFreqMap:
                return self.rawFreqMap[i]/self._totalFreq
    
    def getRawFrequencyOfState(self,i):
        if self.rawFreqMap:
            if i in self.rawFreqMap:
                return self.rawFreqMap[i]


''' Phylogenetic tree structure encapsulation; allow rearrangement of said structure. Tree rearrangements inducing other topologies include Nearest Neighbor Interchange (NNI), Subtree Pruning and Regrafting (SPR), and Tree Bisection and Reconstruction (TBR). Each of these describe a transfer of one node in phylogenetic trees from one parent of a tree to a new parent. Respectively, these operators describe transformations that are subsets of those possible by the successive operator. For example, an NNI operator can perform transformations that are a subset of the transformations possible by the SPR operator. '''

# Date:   Feb 10 2014
# Author: Alex Safatli
# E-mail: safatli@cs.dal.ca

import newick, tree, base

# Exception Handling

class RearrangementError(Exception):
    def __init__(self,val): self.value = val
    def __str__(self): return repr(self.value)

# Movement Type Definitions

TYPE_SPR, TYPE_NNI, TYPE_TBR = 1, 2, 3

# Simulate deep copying.

def dup(topo,where=None):
    if where: new = topology(toLeaf=where)
    else: new = topology()
    new.fromNewick(topo.toNewick())
    if where:
        fake = tree.bipartition(topo,new.fakebranch)
        # Retain locks.
        for lock in new.locked:
            if lock != fake: new.locked.append(lock)      
    return new

# Rearrangement structure.

class rearrangement:
    
    ''' Encapsulates a single rearrangement move
    of type SPR, NNI, ... '''
    
    def __init__(self,struct,type,targ,dest):
        
        ''' Initialize by providing a pointer to
        a base topology, a target branch 
        to be moved, and its destination. '''
        
        self.topol       = struct
        self.type        = type
        self.target      = targ
        self.destination = dest    
    
    def getType(self):
        
        ''' Get the type of movement. '''
        
        if (self.isSPR()):   type = 'SPR'
        elif (self.isNNI()): type = 'NNI'
        elif (self.isTBR()): type = 'TBR'
        else:                type = 'UNKNOWN'
        return type
    
    def isNNI(self): return (self.type == TYPE_NNI)
    def isSPR(self): return (self.type == TYPE_SPR)
    def isTBR(self): return (self.type == TYPE_TBR)
    
    def toTopology(self):
        
        ''' Commit the actual move and return the
        topology. '''
        
        return self.topol.move(self.target,self.destination)
    
    def toNewick(self):
        
        ''' Commit the move but do not create a 
        new structure. Only retrieve resultant
        Newick string; will be more efficient.'''
        
        return self.topol.move(self.target,self.destination,
                               returnStruct=False)
    
    def toTree(self):
        
        ''' Commit the move and transform to tree object. '''
        
        out = tree.tree(self.toNewick())
        out.origin = self.getType()
        return out
    
    def doMove(self): return self.getTopology()
    
    def __str__(self):
        
        t = self.getType()
        return '<%s> move [%s] to [%s]' % (t,self.target,self.destination)

# Tree structure.

class topology(base.treeStructure):
    
    ''' Encapsulate a tree topology, wrapping the newick 
    tree structure. Is immutable. '''
    
    def __init__(self,t=None,rerootToLeaf=True,toLeaf=None):

        ''' Initialize structure with a top-level internal 
        node OR nothing. '''
        
        self.root        = t
        self.forbidden  = {}
        self.branches   = []
        self.locked     = []
        self.fakebranch = None
        self.rerootFlag = rerootToLeaf
        self.rerootLoc  = toLeaf
        
        if t != None: 
            if rerootToLeaf: self.rerootToLeaf(toLeaf)
            self._getAllBranches()
            self._getForbiddenStates()
            self._clearInteriorNodeNames()
            if rerootToLeaf: self._lockLeafBranch()       
       
    def _getAllBranches(self):
        
        ''' PRIVATE: Populate branch list. '''
        
        for br in self.root.children:
            self.branches.append(br)
            self.branches.extend(newick.getAllBranches(br))
        self.partitions = [[x for x in self.branches]]  
       
    def _getForbiddenStates(self):
        
        ''' PRIVATE: Construct forbidden state lists for 
        each branch. '''
        
        # Set up list for each branch.
        for br in self.branches:
            # Anything inside the subtree for that branch
            # is forbidden to be moved to by this branch.
            self.forbidden[br] = [_ for _ in newick.getAllBranches(br)]
            # Any siblings are also forbidden 
            # (give the same tree).
            siblings = br.parent.children
            siblings = [_ for _ in siblings if _!=br]
            self.forbidden[br].extend(siblings)
            # And the parent is forbidden.
            self.forbidden[br].append(br.parent.parent)
        
    def _clearInteriorNodeNames(self):
        
        ''' PRIVATE: Clear all interior node names 
        but store them in case are needed. '''
        
        # Make sure interior nodes are unnamed.
        nodes = self.getAllNodes()
        for n in nodes:
            if n != self.root and len(n.children) != 0 and n.label != '':
                n._label = n.label
                n.label = '' 

    def _lockLeafBranch(self):
        
        ''' PRIVATE: Lock the branch hanging 
        off the root with a leaf. '''
        
        # Get this branch.
        lbranch = None
        for f in self.root.children:
            if (len(f.child.children) == 0):
                lbranch = f
        if (lbranch == None):
            raise RearrangementError(
                'Could not locate leaf branch off root to lock.')
        self.lockBranch(lbranch)

    def _getPartition(self,b):
        
        ''' PRIVATE: Acquire the list of branches: those this particular
        branch could move to without violation of any locks 
        and itself. '''
        
        possible = [x for x in self.branches]
        for lo in self.locked:
            l,r = lo.getBranchListRepresentation()
            if b in l: o = r
            else:      o = l
            for i in o:
                if i in possible: possible.remove(i)
        return possible

    def _flipOperation(self,opname,br):
        
        ''' PRIVATE: Try a particular rearrangement operation on a flipped
        version of the tree. '''

        # Check what bipartition is represented.
        bipart = tree.bipartition(self,br)
        # Get a leaf that can be rerooted to.
        newleaf = None
        leaves  = base.treeStructure.leaves(br.child)
        curleaf = sorted(
            self.getAllLeaves(),key=lambda d: d.label)[0]
        for leaf in leaves:
            if leaf != curleaf:
                newleaf = leaf
                break
        if (newleaf == None): return []
        # Make flipped tree structure.
        n = dup(self,newleaf)
        # Get corresponding branch.
        b = n.getBranchFromBipartition(bipart)
        # Do possible moves from the other way.
        op = getattr(n,opname)
        return op(b,flip=False)

    def rerootToLeaf(self,toleaf=None):
        
        ''' PRIVATE: Reroots the given tree structure such
        that it is rooted nearest the lowest-order leaf. '''
        
        # Determine lowest-order leaf.
        if not toleaf:
            toleaf = sorted(self.getAllLeaves(),
                            key=lambda d: d.label)[0]
        else:
            # Find it in current topology.
            found = False
            for leaf in self.getAllLeaves():
                if leaf.label == toleaf.label:
                    toleaf = leaf
                    found = True
                    break
            if (not found): raise RearrangementError(
                'While rerooting, could not find leaf to reroot to.')

        # Acquire branch connecting to leaf + adjacent node.
        branch  = toleaf.parent
        closest = branch.parent
        
        # Error checking: either of these two do not exist?
        if not (branch):  raise RearrangementError(
            'While rerooting, leaf has no connecting branch.')
        if not (closest): raise RearrangementError(
            'While rerooting, connecting branch of leaf has no parent.')

        # Flip all of the directionality to this node.
        if closest != self.root: 
            t = newick.invertAlongPathToNode(closest,self.root)
            if not t: raise RearrangementError(
                'While rerooting, directions could not be reversed.')
        
        # Strip the lowest node off of the tree.
        closest.children.remove(branch)
        
        # Create a fake branch.
        fakebr = newick.branch(closest,0.0)
        closest.parent = fakebr
        
        # Create a new node to act as the new root.
        root = newick.node(children=[branch,fakebr])
        branch.parent = root
        fakebr.parent = root
        
        # Reassign top-level node.
        self.root = root       
        
        # Remove unary internal nodes.
        newick.removeUnaryInternalNodes(self.root)
        
        # Ensure fake branch is recognized.
        for f in root.children:
            if (f.branch_length == 0 and len(
                f.child.children) > 0):
                fakebr = f
        self.fakebranch = fakebr      
    
    def getBranches(self): return self.branches
    def getLeaves(self): return self.getAllLeaves()
    
    def getBipartitions(self):
        
        ''' Get all bipartitions. '''
        
        bili = []
        br = self.branches
        for b in br:
            bi = tree.bipartition(self,b)
            if not bi in bili: bili.append(bi)
        return bili
    
    def getStrBipartitionFromBranch(self,br):
        
        ''' Given a branch, return corresponding 
        bipartition. '''
        
        right  = base.treeStructure.leaves(br.child)
        others = self.getAllLeaves()
        r      = [x.label for x in right]
        l      = [x.label for x in others 
                  if not x.label in r]
        return l,r
    
    def getBranchFromStrBipartition(self,bip):
        
        ''' Given a bipartition of taxa, return a 
        branch that creates that partition of tree
        taxa. '''
        
        l,r = sorted(bip[0]),sorted(bip[1])
        
        for br in self.branches:
            right  = base.treeStructure.leaves(br.child)
            names  = sorted([x.label for x in right])
            if (names == l) or (names == r):
                return br
        return None
    
    def getBranchFromBipartition(self,bip):
        
        ''' Given a bipartition object, return
        a branch that creates that partition of
        taxa. '''
        
        return self.getBranchFromStrBipartition(
            bip.getStringRepresentation())

    def lockBranch(self,branch):
        
        ''' Given a branch, lock it such that no 
        transitions can ever occur across it. '''
        
        # Transform to bipartition object.
        bipart = tree.bipartition(self,branch)
        
        # Has already been locked?
        if bipart in self.locked: return True
            
        # Is it even in this topology?
        isintopol = (branch in self.branches)
        if (not isintopol): return False
        self.locked.append(bipart)        
        return True

    def move(self,branch,destination,returnStruct=True):
        
        ''' Move a branch and attach to a destination 
        branch. Return new structure, or return merely 
        the resultant Newick string. '''
        
        # Cannot move to these.
        forbidden = self.forbidden[branch]
        if destination in forbidden:
            raise RearrangementError(
                'Cannot move into subtree or to sibling.')
        
        # Check partitions.
        partition = self._getPartition(branch)
        if destination not in partition:
            raise RearrangementError(
                'Cannot move outside of locked partition.')
        
        # Get parents.
        s_parent = branch.parent       # Source parent. 
        t_parent = destination.parent  # Target parent.
        
        # If parents were found, do transform.
        if s_parent != None and t_parent != None:
            
            # Remove the branch.
            s_parent.children.remove(branch)
            
            # Create new node.
            node = newick.node('',[branch])
            
            # Break destination branch in half, attach.
            half  = destination.branch_length/2.0
            outer = newick.branch(destination.child,half,node)
            inner = newick.branch(node,half,t_parent)
            t_parent.children.remove(destination)
            t_parent.children.append(inner)
            node.children.append(outer)
            destination.child.parent = outer
            node.parent = inner
            
            # Check degree of source parent; combine edges if necessary.
            if len(s_parent.children) == 1:
                a           = s_parent.children[0]
                b           = s_parent.parent
                end         = a.child
                start       = b.parent
                comb        = a.branch_length + b.branch_length
                fakebr      = newick.branch(end,comb,start)
                end.parent  = fakebr
                start.children.remove(b)
                start.children.append(fakebr)
            
            if returnStruct:
                # Immutable so recreate new structure.
                result = dup(self)
            else: result = self.toNewick()
            
            # Undo changes.
            s_parent.children.append(branch)
            if len(s_parent.children) == 2:
                end.parent  = a
                start.children.append(b)
                start.children.remove(fakebr)
            t_parent.children.append(destination)
            t_parent.children.remove(inner)
            destination.child.parent = destination

            return result
        
        else: return None
        
    def SPR(self,branch,destination):
        
        ''' Perform an SPR move of a branch to a destination 
        branch, creating a new node there. Returns a 
        rearrangement structure (not the actual new structure) 
        that can then be polled for the actual move; this 
        is in order to save memory. '''
        
        return rearrangement(self,TYPE_SPR,branch,destination)
    
    def NNI(self,branch,destination):
        
        ''' Perform an NNI move of a branch to a destination, 
        only if that destination branch is a parent's parent 
        or a parent's sibling. Returns a rearrangement structure
        (not the actual new structure) that can then be polled
        for the actual move; this is in order to save memory. '''
        
        if newick.isSibling(branch.parent,destination) or \
           (branch.parent and destination == 
            branch.parent.parent):
            return rearrangement(self,TYPE_NNI,branch,destination)
        else: return None
    
    def iterSPRForBranch(self,br,flip=True):    
        
        ''' Consider all valid SPR moves for a given branch 
        in the topology and yield all possible rearrangements
        as a generator. '''    
    
        # Go through possible moves.
        partition = self._getPartition(br)
        if (partition == None):
            raise RearrangementError(
                'Branch not in any partition of topology.')
        forbidden = self.forbidden[br]
        possible  = [x for x in partition if not
                     x in forbidden and not
                     x == br]
        
        # Pass rearrangement structure as yielded object.
        for dest in possible:
            move = self.SPR(br,dest)
            if (move): yield move

        # Flip tree around and try other way.
        if flip:
            ite = self._flipOperation('iterSPRForBranch',br) 
            for it in ite: yield it
    
    def allSPRForBranch(self,br,flip=True):
        
        ''' Consider all valid SPR moves for a given branch 
        in the topology and return all possible rearrangements. '''
        
        return [x for x in self.iterSPRForBranch(br,flip)]
    
    def allSPR(self):
        
        ''' Consider all valid SPR moves for a given topology
        and return all possible rearrangements. '''

        # Output list of structures.
        li = []        
        
        for branch in self.branches:
            li.extend(self.allSPRForBranch(branch))
        
        # Return the list.
        return li

    def iterNNIForBranch(self,br,flip=True):

        ''' Consider all valid NNI moves for a given branch
        in the topology and and yield all possible rearrangements
        as a generator. '''
        
        # Go through possible moves.
        partition = self._getPartition(br)
        if (partition == None):
            raise RearrangementError(
                'Branch not in any partition of topology.')
        possible  = [br.parent.parent] + br.parent.children
        forbidden = self.forbidden[br]
        possible  = [x for x in partition if x in possible and 
                     not x in forbidden and not x == None]
        
        # Pass rearrangement structure as yielded object.
        for dest in possible:
            move = self.NNI(br,dest)
            if (move): yield move

        # Flip tree around and try other way.
        if flip:
            ite = self._flipOperation('iterNNIForBranch',br) 
            for it in ite: yield it        

    def allNNIForBranch(self,br,flip=True):
        
        ''' Consider all valid NNI moves for a given branch 
        in the topology and return all possible rearrangements. '''
        
        return [x for x in self.iterNNIForBranch(br,flip)]

    def allNNI(self):
        
        ''' Consider all valid NNI moves for a given topology
        and return all possible rearrangements. '''

        # Output list of structures.
        li,nni = [],self.allNNIForBranch        
        for branch in self.branches: li.extend(nni(branch))
        return li
    
    def allType(self,type=TYPE_SPR):
    
        ''' Consider all valid moves of a given rearrangement
        operator for a given topology. Uses a given rearrangement 
        operator type defined in this module. For example, calling this 
        function by providing TYPE_NNI as the type will iterate over all 
        NNI operations. By default, the type is TYPE_SPR. '''
        
        if (type == TYPE_SPR):   return self.allSPR()
        elif (type == TYPE_NNI): return self.allNNI()
        else: raise RearrangementError('No rearrangement type of that form is defined.')
    
    def iterTypeForBranch(self,br,type=TYPE_SPR,flip=True):
        
        ''' Iterate over all possible rearrangements for a
        branch using a given rearrangement operator type defined
        in this module. For example, calling this function by providing
        TYPE_NNI as the type will iterate over all NNI operations. By
        default, the type is TYPE_SPR. '''
        
        if (type == TYPE_SPR):   return self.iterSPRForBranch(br,flip)
        elif (type == TYPE_NNI): return self.iterNNIForBranch(br,flip)
        else: raise RearrangementError('No rearrangement type of that form is defined.')
    
    def fromNewick(self,newickstr):
        
        ''' Alias for parse(). '''
        
        return self.parse(newickstr)
    
    def parse(self,newickstr):
        
        ''' Parse a newick string and assign the tree to this
        object. Cannot already be initialized with a tree. '''
    
        if self.root != None:
            raise RearrangementError(
                'Structure already initialized.')
        p = newick.parser(newickstr)
        self.root  = p.parse()
        self.orig = newickstr
        if self.rerootFlag: self.rerootToLeaf(self.rerootLoc)
        self._getAllBranches()
        self._getForbiddenStates()
        self._clearInteriorNodeNames()
        if self.rerootFlag: self._lockLeafBranch()
        
    def toNewick(self):
        
        ''' Return the newick string of the tree. '''
        
        return self.__str__()
    
    def toUnrootedNewick(self):
        
        ''' Return the newick string of the tree as an
        unrooted topology with a multifurcating top-level node. '''
        
        r,a,b = self.root,self.root.children[0],self.root.children[1]
        r.children.remove(a)
        b.child.children.append(a)
        a.parent = b.child
        newic = str(b.child) + ';'
        r.children.append(a)
        b.child.children.remove(a)
        a.parent = self.root
        return newic
        
    def toTree(self):
        
        ''' Return the tree object for this topology. '''
        
        return tree.tree(self.toNewick())
        
    def toUnrootedTree(self):
        
        ''' Return the tree object of the unrooted 
        version of this topology. '''
        
        return tree.tree(self.toUnrootedNewick())
        
    def __str__(self):
        
        ''' Return the newick string of the tree. '''
        
        return str(self.root) + ';'



''' Defines an interface to manage interfacing with the system for respective application calls and implements multiple of these for executables such as FastTree and RAxML. Requires a UNIX environment. '''

# Date:   Oct 16 2014
# Author: Alex Safatli
# E-mail: safatli@cs.dal.ca

import tree
from abc import ABCMeta as abstractclass, abstractmethod
from os import mkdir, getcwd, chdir
from os.path import abspath, isdir, isfile
from subprocess import call, PIPE, Popen as system

# Constants

E_FASTTREE = 'fasttree'
E_RAXML    = 'raxmlHPC'
E_TREEPUZZ = 'puzzle'

# Executable Existence Function

def exeExists(cmd):
    
    ''' Determines whether a function exists in a UNIX environment. '''
    
    return call('type %s'%(cmd),shell=True,stdout=PIPE,stderr=PIPE) == 0

# Temporary Directory Context

class aTemporaryDirectory(object):
    
    ''' A class intended to be used as a context manager that allows
    Python to run in a temporary directory for a finite period of time. '''
    
    def __init__(self,dir=None):
        
        self.path = '/tmp' + dir
        if not isdir(self.path): mkdir(self.path)
        self.current = getcwd()
        
    def __enter__(self):      chdir(self.path)
    def __exit__(self,*args): chdir(self.current)

# Executable Interface

class executable(object):
    
    ''' An interface for the instantation and running of a single instance for a given application. '''

    exeName = None
    __metaclass__ = abstractclass
    
    @abstractmethod
    def getInstructionString(self): pass
    
    def run(self):
        
        ''' Perform a run of this application. '''
        
        if (self.exeName and not exeExists(self.exeName)):
            raise SystemError('%s is not installed on your system.'%(self.exeName))
        sproc = system(self.getInstructionString(),
                       stdout=PIPE,stderr=PIPE,shell=True)
        return sproc.communicate()[0]

# Executable Classes

class treepuzzle(executable):
    
    ''' Wrap TREE-PUZZLE in order to create an intermediate file for CONSEL to read and assign confidence to a set of trees. Requires TREE-PUZZLE to be installed. '''

    exeName = E_TREEPUZZ

    def __init__(self,ali,treefile):
        self.treefile  = treefile
        self.alignment = ali
        if self.alignment == None:
            raise AttributeError('No alignment defined.')
        
    def getInstructionString(self):
        return 'echo "y" | %s %s %s -wsl' % (
            self.exeName,self.alignment.getPhylip(),self.treefile)
    
    def getSiteLikelihoodFile(self):
        self._output = self.run()
        if not isfile('%s.sitelh' % (self.treefile)):
            raise IOError('TREE-PUZZLE did not create site-likelihood output.')
        else: return '%s.sitelh' % (self.treefile) 

class consel(executable):
    
    ''' Denotes a single run of the CONSEL workflow in order to acquire a confidence interval and perform an AU test on a set of trees. Requires CONSEL to be installed. '''
    
    def __init__(self,treeset,alignment,name):
        
        self.treeset   = treeset
        self.name      = name
        self.alignment = alignment
        self.sitelh    = None
        self.raw       = None
        self.rmt       = None
        self.pv        = None
        self.auvals    = list()
        self.interval  = tree.treeSet()
        self.rejected  = tree.treeSet()
        self.instruction = ''
        
    def getInstructionString(self): return self.instruction
    
    def _convertRawData(self):
        
        if not isfile('%s.mt' % (self.name)):
            self.instruction = 'seqmt --puzzle %s %s.mt' % (self.sitelh,self.name)
            self._out = self.run()
            if not isfile('%s.mt' % (self.name)):
                raise IOError('CONSEL seqmt did not create mt file.') 
        self.raw = '%s.mt' % (self.name)
        
    def _createReplicates(self):
        
        if not isfile('%s.rmt' % (self.name)):
            self.instruction = 'makermt %s' % (self.raw)
            self._out = self.run()
            if not isfile('%s.rmt' % (self.name)):
                raise IOError('CONSEL markermt did not create rmt file.')
        self.rmt = '%s.rmt' % (self.name)
        
    def _run(self):
        
        if not isfile('%s.pv' % (self.name)):
            self.instruction = 'consel %s.rmt' % (self.name)
            self._out = self.run()
        self.pv = '%s.pv' % (self.name)
        
    def _getAU(self):
        
        self.instruction = 'catpv %s' % (self.pv)
        pvout = self.out()
        for line in pvout.split('\n'):
            spl = line.split()
            if len(spl) < 3: continue
            elif not spl[2].isdigit(): continue
            it = (int(spl[2])-1,float(spl[4]))
            self.auvals.append(it)
            if it[1] >= 0.05: self.interval.addTree(self.treeset[it[0]])
            else: self.rejected.addTree(self.treeset[it[0]])    
        self._out = pvout
    
    def getInterval(self):
        
        ''' Compute the AU test. Return the interval of trees. '''
        
        with aTemporaryDirectory('/consel__py/'):
            if not isfile('%s.trees.sitelh' % (self.name)):
                self.treefile = self.treeset.toTreeFile(self.name + '.trees',False)
                self.treepuzz = treepuzzle(self.alignment,self.treefile)
                self.sitelh   = self.treepuzz.getSiteLikelihoodFile()
            else: self.sitelh = '%s.trees.sitelh' % (self.name)
            self._convertRawData()
            self._createReplicates()
            self._run()
            self._getAU()
        return self.interval

class fasttree(executable):
    
    ''' Denotes a single run of the FastTree executable in order to acquire an approximate maximum likelihood tree for the input alignment. See http://www.microbesonline.org/fasttree/ for more information on FastTree. Requires FastTree to be installed. '''
    
    exeName = E_FASTTREE
    
    def __init__(self,inp_align,out_file=None,isProtein=True):
        self.alignment = inp_align
        self.isProtein = isProtein
        self.out       = out_file
        
    def getInstructionString(self):
        s = '%s -quiet -nopr < %s' % (self.exeName,self.alignment)
        if self.out: s += ' > %s' % (self.out)
        return s
    
class raxml(executable):
    
    ''' Denotes a single run of the RAxML executable. See http://sco.h-its.org/exelixis/software.html for more information on RAxML. Requires RAxML to be installed. '''
    
    exeName = E_RAXML
    
    def __init__(self,inp_align,out_file,model=None,
                 is_Protein=True,interTrees=False,
                 alg=None,startingTree=None,rapid=False,
                 slow=False,optimizeBootstrap=False,numboot=100,
                 log=None,wdir=None):
        self.alignment     = inp_align
        self.out           = out_file
        self.model         = model
        self.protein       = is_Protein
        self.intermediates = interTrees
        self.alg           = alg
        self.startingTree  = startingTree
        self.slowBoot      = slow
        self.rapidBoot     = rapid
        self.numBoot       = numboot
        self.optBoot       = optimizeBootstrap
        self.log           = log
        self.workdir       = abspath(wdir)
        if not self.model:
            if is_Protein: self.model = 'PROTGAMMAJTT'
            else:          self.model = 'GTRGAMMA'
            
    def getInstructionString(self):
        s = '%s -s %s -n %s -m %s' % (self.exeName,self.alignment,self.out,self.model)
        if self.alg: s += ' -f %s' % (self.alg)
        if self.intermediates: s += ' -j'
        if self.startingTree: s += ' -t %s' % (self.startingTree)
        if self.rapidBoot or self.slowBoot:
            if self.rapidBoot:  s += ' -x 234534251 -N %d' % (self.numBoot)
            elif self.slowBoot: s += ' -b 234534251 -N %d' % (self.numBoot)
        if self.optBoot: s += ' -k'
        if self.workdir: s += ' -w %s' % (self.workdir)
        if self.log: s += ' > %s' % (self.log)
        return s
    
    def runFunction(self,alg):
        self.alg = alg
        self.run()




''' Serialize a phylogenetic landscape into an SQLlite database file made up of three components: all tree IDs and respective scores, the alignment file as a set of sequences, and a representation of the graph as an edge list. '''

# Date:   Apr 9 2014
# Author: Alex Safatli
# E-mail: safatli@cs.dal.ca

from __version__ import VERSION
from alignment import phylipFriendlyAlignment as alignment
from landscape import landscape
from database import SQLiteDatabase
import os, tree

class landscapeWriter(object):

    ''' Encapsulate the writing of a landscape
    to a file format. '''

    def __init__(self, landscape, name):
        
        # Fields
        self.landscape = landscape
        self.name      = name
        self.cleankey  = ''
        self.graph     = landscape.graph
        self.database  = None        
        
        # Clean the input landscape name.
        self._cleanName()

    def _cleanName(self):

        for x in self.name:
            if x.isalnum(): self.cleankey += x
            else: self.cleankey += '_'
        self.cleansuff = '%s.landscape' % (self.cleankey)

    def _schema(self):
        
        # Get the database object.
        dbobj = self.database
        
        # Create all content tables.
        dbobj.newTable('alignment',('seqid','integer'),('seqname','text'),('sequence','text'))
        dbobj.newTable('trees',('treeid','integer'),('newick','text'),
                       ('origin','text'),('ml','real'),('pars','real'),('explored','boolean'))
        dbobj.newTable('graph',('source','text'),('origin','text'))
        dbobj.newTable('locks',('treeid','integer'),('branchid','integer'))
        
        # Create a table to store metadata.
        dbobj.newTable('metadata',('key','text'),('value','text'))

    def _dump(self,path='.'):

        # Open the file.
        fpath = os.path.join(path,self.cleansuff)
        if os.path.isfile(fpath): os.unlink(fpath)
        o = SQLiteDatabase(fpath)
        self.database = o
        
        # Construct the schema for the landscape.
        self._schema()
        
        # Include metadata about the landscape.
        o.insertRecord('metadata',['name',self.name])
        o.insertRecord('metadata',['version',VERSION])
        
        # Add the alignment.
        index = 0
        if self.landscape.alignment != None:
            for s in self.landscape.alignment:
                o.insertRecord('alignment',[index,s.name,s.sequence])
                index += 1
        
        # Add all of the trees.
        for i in self.landscape.iterNodes():
            t = self.landscape.getTree(i)
            o.insertRecord('trees',[int(i),t.getNewick(),t.getOrigin(
                ),t.getScore()[0],t.getScore()[1],
                self.landscape.getVertex(i).isExplored()])
        
        # Add the graph in its entirety as its adjacency list.
        adj_list = self.graph.edges_iter()
        for tupl in adj_list:
            o.insertRecord('graph',[int(x) for x in tupl])
        
        # Add all of the locks.
        for l in self.landscape.getLocks():
            tree = l.topology.toTree()
            treeIndex = self.landscape.indexOf(tree)
            o.insertRecord('locks',[treeIndex,l.getBranchIndex()])
        
        # Close the file.
        o.close()

        # Return the path.
        return fpath

    def writeFile(self,path='.'):
        
        ''' Write the landscape serialized file to given path. '''

        return self._dump(path)

class landscapeParser(object):

    ''' Encapsulates the construction of a landscape
    object from a sqlite landscape file. '''

    def __init__(self,path):
        self.file = path
        self.metadata = None
        self.name = None
        self.database = None
        self.alignment = None
        self.trees = []
        self.explored = {}
        self.landscape = None

    def getName(self):
        
        ''' Acquire the name of the parsed landscape. '''
        
        if self.name is None:
            if len(self.metadata) > 0:
                for metadata in self.metadata:
                    if metadata[0] == 'name':
                        self.name = metadata[1]
            else: self.name = os.path.splitext(os.path.basename(self.file))[0]
        return self.name

    def _getLandscapeMetadata(self):
        
        ''' Check metadata in landscape if present. '''
        
        tables = self.database.getTables()
        if 'metadata' in tables:
            return self.database.getRecords('metadata')
        else: return []

    def _makeAlignment(self):
        
        ''' Construct a pseudo-FASTA string to create a new alignment from. '''
        
        pseudofasta = ''
        sequences = self.database.iterRecords('alignment')
        for sequence in sequences:
            _,name,seq = sequence
            if name != '' and seq != '':
                pseudofasta += '>%s\n%s\n' % (name,seq)
        if pseudofasta != '':
            self.alignment = alignment(str(pseudofasta))
        
    def _getTrees(self):
    
        floatIfNotNone = lambda d: float(d) if d != None else d
        intIfNotNone   = lambda d: int(d) if d != None else d
        for t in self.database.iterRecords('trees'):
            id,newick,orig,ml,pars,exp = t
            if newick != '':
                trobj = tree.tree(str(newick))
                trobj.origin = str(orig)
                trobj.name = int(id)
                trobj.score = [floatIfNotNone(ml),intIfNotNone(pars)]
                self.trees.append(trobj)
                self.explored[trobj] = bool(exp)
        self.trees = sorted(self.trees,key=lambda d: d.name)

    def _getGraph(self):
        
        for e in self.database.iterRecords('graph'):
            source,target = e
            self.landscape.graph.add_edge(int(source),int(target))

    def parse(self):
        
        ''' Parse the file. '''

        # Get database object.
        fpath = self.file
        if not os.path.isfile(fpath):
            raise IOError('Landscape file not found.')
        o = SQLiteDatabase(fpath)
        self.database = o
        
        # Get metadata.
        self.metadata = self._getLandscapeMetadata()
        
        # Get the alignment.
        self._makeAlignment()
        
        # Get all of the trees.
        self._getTrees()
        
        # Construct the landscape.
        self.landscape = landscape(self.alignment,starting_tree=self.trees[0],
                                   operator=self.trees[0].getOrigin(),root=False)
        for t in self.trees:
            add = self.landscape.addTree(t)
            self.landscape.getVertex(add).setExplored(self.explored[t])
        
        # Ensure all edges are established.
        self._getGraph()

        return (self.landscape,self.getName())
